__title__ = "PyMatting"
__version__ = "1.1.14"
__author__ = "The PyMatting Developers"
__email__ = "pymatting@gmail.com"
__license__ = "MIT"
__uri__ = "https://pymatting.github.io"
__summary__ = "Python package for alpha matting."
